# frozen_string_literal: true

module Hermes
  class HousingChannel
    include Hermes::Concerns::CommonFieldsConcern

    writer_roles 'Administrator'
    reader_roles 'Administrator'
    owner_field :user_id

    field :key, type: String
    field :label, type: String

    common_indexes *%i[key label]

    before_save :format_key

    validates_presence_of :key
    validates_uniqueness_of :key

    private

    def format_key
      self.key = key.parameterize('_') if key_changed?
    end
  end
end