# frozen_string_literal: true

# Legende
# Les script a plusieurs parties indique que les bash (Batch dans l'app)
# peuvent s'initier avec differents types de source
# Le besoin est donc d'isoler la logique associee au modele actuel
# ne pas oublier d'indexer les champs simples

module Hermes
  # Housing offer model
  class HousingOffer
    include Hermes::Concerns::CommonFieldsConcern
    include Hermes::Concerns::AuditableConcern
    include Hermes::Concerns::AlsIdConcern
    include Hermes::Concerns::Status
    include Hermes::Concerns::ReservationHolder
    include ::Concerns::Geolocalizable
    include Hermes::Concerns::SopraXmlSender
    
    include Mongoid::Locker
    index({locking_name: 1}, name: 'mongoid_locker_index', partial_filter_expression: {locking_name: {:$exists => true}}, expire_after_seconds: lock_timeout)
    
    hermes_tools 'typology'
    
    RENTED_STATUSES = %w[LOUE loué loue PLACE placé place].freeze
    PUBLISHED_STATUSES = ['LOGEMENT PUBLIE', 'PUBLIE'].freeze
    ATTRIBUTION_STATUSES = %w[requalification REQUALIFICATION]
    
    ORIENTATION_BOOLS = %i[
      priority_public
      single_parent
      mutation
    ].freeze
    
    field :requalification, type: Boolean, default: false
    field :dalo, type: Boolean
    field :entry_date, type: DateTime
    field :external_ref, type: String
    field :communicate_exact_address, type: Boolean
    field :typology, type: String
    field :type, type: String
    field :age, type: String
    field :department, type: String
    field :publication_end_date, type: DateTime
    field :max_applicants, type: Integer
    field :minimal_publication_days, type: Integer, default: 3
    field :rent_amount, type: Float
    field :rent_with_charges, type: Float
    field :surface, type: Integer
    field :rooms, type: Integer
    field :exclusivity_end_date, type: DateTime
    field :nb_of_eligible_employees, type: Integer
    field :nb_of_requests_from_employees, type: Integer
    field :availability_date, type: DateTime
    field :description, type: String
    field :details, type: String
    field :effort_rate, type: Float
    field :min_effort_treshold, type: Float
    field :single_parent, type: Boolean
    field :reserved, type: Boolean
    field :reservation_agreement_ref, type: String
    field :priority_public, type: Boolean
    field :mobility, type: Boolean
    field :bedrooms, type: Integer
    field :bathrooms, type: Integer
    field :floor, type: Integer
    field :floors, type: Integer
    field :elevators, type: Integer
    field :has_elevator, type: Boolean
    field :has_gardian, type: Boolean
    field :outside, type: String
    field :cellars, type: Integer
    field :has_cellar, type: Boolean
    field :parking_type, type: String
    field :parking_charges, type: Float
    field :pmr, type: Boolean
    field :kitchen_type, type: String
    field :construction_year, type: Date
    field :address, type: String
    field :address_complements, type: String
    field :mutation, type: Boolean
    field :eligibility_query, type: String
    field :threshold_type, type: String
    field :threshold_subtype, type: String
    field :heating_type, type: String
    field :hot_water_distribution, type: String
    field :individual_heating, type: Boolean
    field :date_publication_start, type: DateTime
    field :date_publication_end, type: DateTime
    field :kind, type: String
    field :building_type, type: String
    field :already_rent, type: Boolean
    field :guarantee_deposit, type: Float
    field :rental_charges, type: Float
    field :reference, type: String
    field :release_date, type: Date
    field :dpe_ges, type: String
    field :dpe_conso, type: String
    
    field :ceil_category_1, type: Float
    field :ceil_category_2, type: Float
    field :ceil_category_3, type: Float
    field :ceil_category_4, type: Float
    field :ceil_category_5, type: Float
    field :ceil_category_6, type: Float
    field :ceil_category_7, type: Float
    field :ceil_category_8, type: Float
    field :ceil_category_9, type: Float
    field :ceil_category_10, type: Float
    field :ceil_category_11, type: Float
    field :ceil_category_12, type: Float
    field :floor_category_1, type: Float
    field :floor_category_2, type: Float
    field :floor_category_3, type: Float
    field :floor_category_4, type: Float
    field :floor_category_5, type: Float
    field :floor_category_6, type: Float
    field :floor_category_7, type: Float
    field :floor_category_8, type: Float
    field :floor_category_9, type: Float
    field :floor_category_10, type: Float
    field :floor_category_11, type: Float
    field :floor_category_12, type: Float
    field :min_revenues, type: Float
    field :total_charges, type: Float
    field :min_effort_rate, type: Float, default: 0.333
    field :tense_area, type: Boolean
    field :subsidiary_company, type: String
    field :rpls_id, type: String
    field :operation_name, type: String
    field :administrator, type: String
    field :administration_area, type: String
    field :potential_request_name_town, type: String
    field :potential_request_name_county, type: String
    field :number_of_companies_30_minutes_around, type: Integer
    field :construction_type, type: String
    field :address_can_be_shared, type: String
    field :qpv, type: Boolean
    field :total_floors, type: String
    field :address_extended, type: String
    field :test_created_at, type: DateTime
    field :already_inhabited, type: Boolean
    field :new_offer, type: Boolean
    field :lessor_name, type: String
    field :lessor_siren, type: String
    field :als_siren, type: String
    field :postal_code, type: String
    field :insee_code, type: String
    field :residence_code, type: String
    field :reserving_business_name, type: String
    field :reserving_siren, type: String
    field :residence_title, type: String
    # field :main_picture_id, type: BSON::ObjectId
    # field :pictures, type: String
    field :agency_id, type: BSON::ObjectId
    field :agency_name, type: String
    field :lessor_contact_name, type: String
    field :lessor_email, type: String
    field :lessor_phone_number, type: String
    field :rent_without_charges, type: String
    field :terraces_or_balconies, type: Integer
    field :gardens, type: Integer
    field :has_garden, type: Boolean
    field :green_area, type: Boolean
    field :condo, type: Boolean
    field :furnished_lodging, type: Boolean
    field :virtual_visit, type: Boolean
    field :duplex, type: Boolean
    # TODO : verifier le match au noveau du WIKI
    field :first_quartile, type: String
    field :other_quartiles, type: String
    field :lessor_convention_reference, type: String
    field :reserving_agency_name, type: String
    field :reserving_agency_siren, type: String
    field :is_public, type: Boolean
    field :published, type: Boolean
    field :accept_low_occupation, type: Boolean, default: false
    field :designation_date_end, type: DateTime
    field :designation_period_duration, type: Integer
    field :lessor_orientation, type: String
    field :disability_accesses, type: Integer
    field :has_disability_access, type: Boolean
    field :eligible_requests_nb, type: Integer
    field :applicated_nb, type: Integer
    field :entities_siret, type: Array
    field :eligible_requests_nb_detailed, type: Hash # of form { priority_public: 23, mutation: 45, ... }
    field :commercialized, type: Boolean, default: false
    field :entity_ids, type: Array, default: []
    # associated pictures
    field :ordered_picture_ids, type: Array
    
    field :is_mad, type: Boolean, default: false
    field :is_mad_updated_at, type: DateTime
    field :mad_als_date, type: DateTime
    
    field :max_family_size, type: Integer, default: 0
    field :couple_min_family_size, type: Integer, default: 0
    field :single_min_family_size, type: Integer, default: 0
    field :flux_5_sended, type: Boolean
    field :last_update_by_lessor, type: DateTime
    field :auto_destroy_next_day, type: Boolean
    
    
    # validates :entities_siret_length
    validates :lessor_siren, length: {is: 9}, allow_blank: true
    validates :als_siren, length: {is: 9}, allow_blank: true
    
    validates :reserving_siren, length: {is: 9}, allow_blank: true
    validates :reserving_agency_siren, length: {is: 9}, allow_blank: true
    
    
    alias_attribute :district, :city
    
    common_indexes *%i[
      entry_date
      external_ref
      typology
      lessor_status
      age
      publication_end_date
      max_applicants
      minimal_publication_days
      rent_amount
      rent_with_charges
      exclusivity_end_date
      availability_date
      min_effort_rate
      min_effort_treshold
      single_parent
      reserved
      priority_public
      pmr
      address
      mutation
      threshold_type
      threshold_subtype
      individual_heating
      date_publication_start
      date_publication_end
      building_type
      ceil_category_1
      ceil_category_2
      ceil_category_3
      ceil_category_4
      ceil_category_5
      ceil_category_6
      ceil_category_7
      ceil_category_8
      ceil_category_9
      ceil_category_10
      ceil_category_11
      ceil_category_12
      floor_category_1
      floor_category_2
      floor_category_3
      floor_category_4
      floor_category_5
      floor_category_6
      floor_category_7
      floor_category_8
      floor_category_9
      floor_category_10
      floor_category_11
      floor_category_12
      min_revenues
      min_effort_rate
      operation_name
      address_can_be_shared
      latitude
      longitude
      qpv
      als_siren
      postal_code
      insee_code
      agency_id
      eligible_requests_nb
      applicated_nb
      lessor_siren
      lessor_status
      lessor_status_id
      lessor_status_updated_at
      old_lessor_status
      offer_status
      offer_status_id
      offer_status_updated_at
      old_offer_status
      is_mad
      is_mad_updated_at
      commercialized
      is_public
      published
      last_update_by_lessor
      lessor_id
      target_objective_ids
      auto_destroy_next_day
    ]
    
    attr_accessor :will_update_lessor_objectives, :minimal_publication_date
    
    belongs_to :threshold, class_name: 'Hermes::Threshold'
    # TODO: handle also with_entity_ids=
    has_many :housing_entity_offers, class_name: 'Hermes::HousingEntityOffer'
    belongs_to :lessor, class_name: 'Hermes::Lessor'
    has_and_belongs_to_many :target_objectives, class_name: 'Hermes::LessorObjective'
    has_many :scores, class_name: 'Hermes::Score'
    field :picture_ids_was, type: Array
    has_many :pictures, class_name: 'Hermes::Picture'
    
    writer_roles 'Superadmin', 'Administrator'
    reader_roles 'Manager'
    owner_field :user_id
    
    reader_auth_roles 'collaborateur offre logement'
    
    before_save :before_save_actions # https://dev.hermes.openstack.local/admin/families/5cffb7af4ac06e63bbd93298/mapping
    after_create :after_create_actions
    after_save :after_save_actions
    after_update :after_update_actions
    
    ELIGIBILITY_FIELDS = %w[
      entity_ids
      mutation
      mobility
      offer_status
      priority_public
      single_parent
      rent_with_charges
      is_public
      single_min_family_size
      couple_min_family_size
      max_family_size
      min_revenues
      department
      threshold_type
      threshold_subtype
    ].concat(
      (1..12).each_with_object([]) do |i, arr|
        %w[ceil floor].each {|type| arr << "#{type}_category_#{i}"}
      end
    ).freeze
    
    THRESHOLDS_FIELDS = %w[
      threshold_type
      threshold_subtype
    ].freeze
    
    HousingOffer.add_status(:single, 'lessor_status')
    HousingOffer.add_status(:single, 'offer_status')
    
    scope :will_trigger_objective, ->(objective_id) {
      where(target_objective_ids: objective_id, :lessor_status.nin => RENTED_STATUSES, :offer_status.nin => ATTRIBUTION_STATUSES)
    }
    
    ###############################
    #### Notification #############
    include ::Concerns::NotificationCenter::Notifiable
    
    def self.targets_selector(housing_offer, opt = {})
      applicant_ids = Hermes::Score.where(housing_offer_id: housing_offer.id).map(&:applicant_id)
      Hermes::Applicant.where(:_id.in => applicant_ids || [])
    end
    
    def notifiable_title(key)
      I18n.t("#{key}.title", title: id)
    end
    
    def notifiable_body(key)
      I18n.t("#{key}.body")
    end
    
    def reset_siret_siren
      off.lessor_siren = off.lessor_siren&.slice(0, 9)
      off.als_siren = off.als_siren&.slice(0, 9)
      off.reserving_siren = off.reserving_siren&.slice(0, 9)
      off.reserving_agency_siren = off.reserving_agency_siren&.slice(0, 9)
    
    end
    
    def notifiable_path
      "https://als-hermes-int.be-idys.com/housing_offers/#{id}"
    end
    
    ###############################
    ###############################
    
    def construction_year=(string)
      super(Date.new(Integer(string)))
    end
    
    def contact_group
      lessor&.contact_group
    end
    
    def main_contact
      contact_group&.main_contact
    end
    
    def entities
      Hermes::Entity.in(id: housing_entity_offers.pluck(:entity_id))
    end
    
    def entity_ids
      entities.pluck(:id)
    end
    
    def entity_ids=(*arr)
      previous_ids = entities.pluck(:id).map(&:to_s)
      new_ids = (arr || []).map(&:to_s)
      (new_ids - previous_ids).each do |entity_id|
        Hermes::HousingEntityOffer.create(entity_id: entity_id, housing_offer_id: id)
      end
      Hermes::HousingEntityOffer.where(:entity_id.in => (previous_ids - new_ids), housing_offer: id).destroy
      super(entity_ids)
    end
    
    def lessor_id=(lessor_new_id)
      reutrn unless lessor_id != id
      
      Hermes::HousingEntityOffer.find_by(entity_id: lessor_id, housing_offer_id: id)&.destroy
      Hermes::HousingEntityOffer.create(entity_id: lessor_new_id, housing_offer_id: id)
      super(lessor_new_id)
    end
    
    def agency
      Hermes::Entity.find_by(id: agency_id)
    end
    
    def history
      Hermes::LessorObjectiveHistory.where(housing_offer_id: self.id)
    end
    
    def detailed_score
      Hermes::Score.where(housing_offer_id: self.id)
    end
    
    def inherit_lesssor_data
      return unless lessor
      
      self.lessor_name = lessor.title
    end
    
    def update_extras_presence
      %w[elevator garden cellar disability_access].each do |extra_name|
        send("has_#{extra_name}=", send(extra_name.pluralize)&.send('>', 0))
      end
    end
    
    # set relationship with threshold
    def update_own_thresholds
      return unless threshold
      
      threshold.set_offer_thresholds(self)
    end
    
    def update_eligbility_old
      # https://dev.hermes.openstack.local/admin/batchs/5d2301804ac06e2ce9126715/script
      # script tri-partie (offer || (applicant || requester))
      if target = (filterer = FilterGenerator.new(context.select {|key| %w[offer_id applicant_id request_id].include? key})).target_instance
        new_sfs = target.superfields
        new_sfs[filterer.process_watch_field] = 'off'
        if target.set(superfields: new_sfs) && filterer.set_filter_on_target
          Rails.logger.info "\n BASH process ended with filters #{filterer.filters} at #{target.reload.superfields[filterer.process_watch_field]}\n\n"
          "Elligibility on #{filterer.target} passed with result : #{filterer.filters}"
        else
          false
        end
      else
        Rails.logger.error "Element #{context.each_with_object([]) do |(_key, value), arr|
          arr.push value if value.present?
        end.first} could not be found"
        false
      end
    end
    
    # integration olivier => batch https://dev.hermes.openstack.local/admin/families/5cffb7af4ac06e63bbd93298/mapping-
    def update_family_size
      surface = self.surface || 0
      typologint = case typology
                   when /^T([1-9][0-9]*?)$/
                     Integer(::Regexp.last_match(1))
                   when 'Chambre'
                     1
                   when 'T6 et plus'
                     rooms
                   end
      typologint = (typologint&.positive? ? typologint : rooms) || 0
      min_family_size = (min_calculated = typologint - 2) < 1 ? 1 : min_calculated
      max_size = typologint < 6 ? (typologint + 1).to_s : 24
      single_min_dependents = 0
      i = 0
      surface_bis = surface || 0
      while surface_bis.positive?
        can_increment = surface_bis -= if i == 1
                                         7
                                       elsif i < 8
                                         9
                                       else
                                         8
                                       end
        single_min_dependents += if (i.zero? && can_increment > 6) || (i.between?(1, 7) && can_increment > 8) || (i > 7 && can_increment > 7)
                                   1
                                 elsif i.zero? && can_increment.negative?
                                   -1
                                 else
                                   0
                                 end
        i += 1
      end
      s_min_family_size = [min_family_size, single_min_dependents].min - ((accept_low_occupation == true ? 1 : 0) * 1)
      couple_min_family_size = ((surface - 16) / 9).floor + 2 - ((accept_low_occupation === true ? 1 : 0) * 1)
      c_min_family_size = [min_family_size, couple_min_family_size].min - ((accept_low_occupation == true ? 1 : 0) * 1)
      int_max_size = max_size.is_a?(String) ? Integer(max_size.sub('+', '')) : max_size
      self.single_min_family_size = s_min_family_size > int_max_size ? int_max_size : s_min_family_size
      self.couple_min_family_size = c_min_family_size > int_max_size ? int_max_size : c_min_family_size
      self.max_family_size = max_size
    end
    
    def update_individual_heating
      self.individual_heating = heating_type && heating_type =~ /^Individuel/
    end
    
    def update_charges
      tc = (rental_charges || 0.0) + (parking_charges || 0.0)
      if individual_heating && surface && heating_type =~ /électrique$/
        tc += if individual_heating
                if electric
                  1.3 * surface
                else
                  1.5 * surface
                end
              else
                0
              end
      end
      self.total_charges = tc
    end
    
    def update_min_revenues
      formated_min_effort_rate = (min_effort_rate&.to_f || 0.0)
      self.min_revenues = (formated_min_effort_rate.positive? ? 1 / formated_min_effort_rate : 0.0) * (rent_with_charges || 0.0)
    end
    
    def update_rent_with_charges
      self.rent_with_charges = (rent_amount || 0.0) + (total_charges || 0.0)
    end
    
    def update_threshold
      self.threshold = Hermes::Threshold.find_by("$or" => [{type: threshold_type}, {codes: threshold_type}])
    end
    
    def formated_eligibility
      JSON.parse(eligibility_query).each_with_object({}) do |(key, val), h|
        h[key.to_sym] = val unless val.nil?
      end
    end
    
    def calculate_eligible_requests_nb
      return if is_public
      
      query = formated_eligibility.merge(is_active: true)
      # only = %i[is_active].concat(ELIGIBILITY_FIELDS.map(&:to_sym))
      # TODO: using only didn't worked out, but can becomme an optimization
      # Hermes::HousingRequest.only(only).where(query).count
      Hermes::HousingRequest.where(query).count
    end
    
    def update_eligible_requests_nb_detail
      self.eligible_requests_nb_detailed ||= {}
      ORIENTATION_BOOLS.each do |orientation_name|
        orientation = send(orientation_name)
        next unless orientation
        
        self.eligible_requests_nb_detailed[orientation_name] = eligible_requests.where(orientation_name.to_sym => orientation).count
      end
      
      if reserved && entity_ids
        self.eligible_requests_nb_detailed[entities] = entity_ids.each_with_object({}) do |entity_id, h|
          h[entity_id] = eligible_requests.where(entity_id: entity_id).count
        end
      end
    end
    
    def update_eligible_requests_nb
      query = formated_eligibility.merge(is_active: true)
      # only = %i[is_active].concat(ELIGIBILITY_FIELDS.map(&:to_sym))
      # TODO: using only didn't worked out, but can becomme an optimization
      # Hermes::HousingRequest.only(only).where(query).count
      eligible_requests = Hermes::HousingRequest.where(query)
      
      self.eligible_requests_nb = eligible_requests.count
    end
    
    def calculate_applicated_nb
      query = formated_eligibility.merge(is_active: true, housing_offer_id: id)
      # only = %i[is_active housing_offer_id].concat(ELIGIBILITY_FIELDS)
      # TODO: using only didn't worked out, but can becomme an optimization
      # Hermes::HousingRequest.only(only).where(query).count
      Hermes::HousingRequest.where(query).count
    end
    
    def inc_applicated_nb(int)
      self.applicated_nb ||= 0
      self.applicated_nb += int
    end
    
    def update_applicated_nb
      self.applicated_nb = calculate_applicated_nb
    end
    
    def update_quantity
      update_applicated_nb
      update_eligible_requests_nb
    end
    
    def update_department
      self.department = (
      begin
        postal_code[0..1].to_s
      rescue StandardError
        nil
      end)
    end
    
    def update_is_mad_updated_at
      self.designation_date_end = DateTime.current + 30.days
      self.is_mad_updated_at = DateTime.current
    end
    
    def handle_unpublished_status
      if PUBLISHED_STATUSES.include?(offer_status) && date_publication_start?
        expiry_days = 7 # default
        setting = Hermes::HousingOfferExpirationSetting.find_by(department: department, lessor_id: lessor_id)
        setting = Hermes::HousingOfferExpirationSetting.find_by(department: department, lessor_id: nil) if setting.nil?
        if setting
          expiry_days = setting.expiry_days
          set(
            max_applicants: setting.max_applicants,
            minimal_publication_days: setting.minimal_publication_days,
          )
        end
        
        @minimal_publication_date ||= date_publication_start + expiry_days
        
        if !max_applicants.nil? && max_applicants <= applicated_nb
          @minimal_publication_date = date_publication_start + minimal_publication_days
          if Time.zone.now > @minimal_publication_date
            set(publication_end_date: Time.zone.now, offer_status: 'COMMERCIALISATION NON PUBLIEE')
            return
          end
        end
        
        set(publication_end_date: @minimal_publication_date)
        UnpublishHousingOfferJob.set(wait_until: @minimal_publication_date).perform_later(self.id.to_s)
      end
    end
    
    def handle_entities_relationships
      
      
      formated_siret = entities_siret || []
      formated_siren = []
      formated_siret.each {|element| formated_siren.push element.slice(0, 9)}
      formated_siret_was = entities_siret_was || []
      formated_siren_was = []
      formated_siret_was.each {|element| formated_siren_was.push element.slice(0, 9)}
      ap formated_siren
      formated_siren_plus_lessor = (formated_siren + [lessor_siren]).compact
      formated_siren_was_plus_lessor = (formated_siren_was + [lessor_siren_was]).compact
      
      all_removed_entities = Hermes::Entity.where(:siren.in => formated_siren_was_plus_lessor - formated_siren_plus_lessor)
      all_added_entities = Hermes::Entity.where(:siren.in => formated_siren_plus_lessor - formated_siren_was_plus_lessor)
      
      not_lessor = {siren: {:$ne => lessor_siren}}
      removed_entities = all_removed_entities.where(not_lessor)
      added_entities = all_added_entities.where(not_lessor)
      Hermes::HousingEntityOffer.where(
        :entity_id.in => removed_entities.pluck(:id),
        housing_offer_id: id
      ).set(
        is_active: false,
        updated_at: DateTime.current
      )
      
      added_entities.each do |entity|
        offer_entity = Hermes::HousingEntityOffer.find_or_create_by(housing_offer_id: id, entity_id: entity.id)
        offer_entity.is_active = true
        offer_entity.save
      end
      
      # TODO: better handling for groups
      all_removed_entities.each do |entity|
        readers.delete("g_#{entity.supervisor_group.id}")
        writers.delete("g_#{entity.manager_group.id}")
      end
      all_added_entities.each do |entity|
        readers.push("g_#{entity.supervisor_group.id}")
        writers.push("g_#{entity.manager_group.id}")
      end
    end
    
    def handle_is_mad
      return if is_mad == true
      
      self.is_mad = (offer_status == 'EN ATTENTE COMMERCIALISATION' && lessor_status == 'EN ATTENTE PUBLICATION')
      update_is_mad_updated_at if is_mad_changed?
    end
    
    def main_picture
      pictures.first
    end
    
    def soft_delete
      # "todo : soft delete of housing_offer"
      
      return if Rails.env == 'test'
      Hermes::BoClient::OfferStatusTask.new(housing_offer_id: id).execute
    end
    
    def set_log_action_history(action)
      Hermes::Log.create(
        code: 'HERMES_HISTORY_LOG',
        action: "HousingOffer #{action}",
        housing_offer_id: self.id,
        current_user_id: Hermes::HousingOffer.current_user_id
      )
      Hermes::HousingOffer.current_user_id = nil
    end
    
    def assign_target_objective
      # self.target_objective_ids=[]
      self.target_objectives = lessor.get_target_objectives_for(self)
    end
    
    def before_save_actions
      if lessor_siren
        self.lessor = Hermes::Lessor.find_by(siren: lessor_siren)
      end
      update_extras_presence
      if entities_siret_changed? || lessor_siren_changed?
        handle_entities_relationships
      end
      inherit_lesssor_data if lessor_id_changed?
      update_department if postal_code_changed?
      update_individual_heating if heating_type_changed?
      update_charges if rental_charges_changed? || parking_charges_changed?
      update_rent_with_charges if rent_amount_changed? || total_charges_changed?
      if rent_with_charges_changed? || min_effort_rate_changed?
        update_min_revenues
      end
      update_family_size if typology_changed? || accept_low_occupation_changed?
      update_threshold if threshold_type_changed?
      if threshold_id_changed? || threshold_subtype_changed?
        update_own_thresholds
      end
      
      if RENTED_STATUSES.include?(lessor_status)
        lessor_status = "LOUE"
      end
      # TODO: handle is_mad for status
      handle_is_mad if lessor_status_changed? || offer_status_changed?
      update_eligbility if (changes.keys & ELIGIBILITY_FIELDS).length.positive?
      update_quantity if eligibility_query_changed?
      target_objective_computation_needed = (agency_id_changed? ||
        lessor_id_changed? ||
        operation_name_changed? ||
        district_changed? ||
        department_changed? ||
        threshold_type_changed? ||
        new_offer_changed? ||
        priority_public_changed? ||
        typology_changed?)
      assign_target_objective if (lessor && target_objective_computation_needed && (!lessor_status_changed? && !RENTED_STATUSES.include?(lessor_status)))
      
      handle_requalification_status
    end
    
    def handle_requalification_status
      if ATTRIBUTION_STATUSES.include?(offer_status)
        self.requalification = true
      end
    end
    
    def update_lessor_objectives
      lessor&.handle_objectives_for(self)
    end
    
    def destroy_tomorrow_if_no_update
      AutoDestroyOfferJob.set(wait: 36.hours).perform_later(id.to_s)
    end
    
    def after_create_actions
      set_log_action_history('created') if Hermes::HousingOffer.current_user_id
    end
    
    def after_save_actions
      check_status_changed
      
      update_lessor_objectives if lessor_status_changed? && RENTED_STATUSES.include?(lessor_status)
      destroy_tomorrow_if_no_update if auto_destroy_next_day && last_update_by_lessor_changed?
      
      soft_delete if offer_status_changed? && offer_status == "RETIRE PAR LE BAILLEUR"
      
      Log.create(code: "HERMES_FLUX5", action: "housing_offer : #{self.id} -- flux 5 aftersave, verif flux_5_sended : " + flux_5_sended.to_s, data: self.to_json.to_s)
      if is_mad_changed? && is_mad && !flux_5_sended
        self.set(flux_5_sended: true, updated_at: DateTime.now)
        Log.create(code: "HERMES_FLUX5", action: "housing_offer : #{self.id} -- flux 5 envoyé", data: self.to_json.to_s)
        Hermes::BoClient::AddOfferTask.new(housing_offer_id: id).execute
      end
      
      handle_unpublished_status
      
      set(published: PUBLISHED_STATUSES.include?(offer_status))
    end
    
    def after_update_actions
      update_affiliated_scores if department_changed?
      set_log_action_history('updated') if Hermes::HousingOffer.current_user_id
    end
    
    def update_affiliated_scores
      ToucheScoresJob.perform_later(
        Hermes::Score.only(:id, :housing_offer_id)
          .where(housing_offer_id: id)
          .pluck(:id)
          .map(&:to_s)
      )
    end
    
    def for_single_parent
      if single_min_family_size > max_family_size
        [max_family_size]
      elsif single_min_family_size != max_family_size
        (single_min_family_size..max_family_size).map(&:to_s)
      else
        [single_min_family_size]
      end
    end
    
    def update_lessor_on_all_offers
      offs = Hermes::HousingOffer.all
      counter = 0
      offs.each do |off|
        if (off.lessor_siren)
          off.lessor = Hermes::Lessor.find_by(siren: off.lessor_siren)
          off.save
          counter = counter+1
          ap counter
        end
      end
    end
    
    def update_eligbility
      # boolean filters
      boolean_filters = %i[
        single_parent
        priority_public
        mutation
      ]
      h = attributes.select do |key, val|
        boolean_filters.include?(key.to_sym) && val
      end.each_with_object({}) do |(key, val), h|
        h[key.to_sym] = val unless val.nil?
      end
      
      # reserved filters reserve field => 'reserved'
      
      if reserved
        entities.each do |entity|
          h[:$and] ||= []
          h[:$and] << {
            '$or' => [{entity_id: entity.id.to_s}, {entity_id: nil}]
          }
        end
      end
      
      # category filter
      h[:$and] ||= []
      h[:$and] << {
        :$or => (1..12).map do |category|
          {
            category_type: category,
            reference_tax_revenues: {
              '$lte': get_tax_threshold(category),
              '$gte': get_tax_threshold(category, true)
            }
          }
        end
      }
      
      # typology filter
      single_min = single_min_family_size
      couple_min = couple_min_family_size
      max_size = max_family_size
      if max_size
        max_size = Integer(max_size =~ /\+$/ ? max_size.sub('+', '') : max_size)
      end
      single_min = single_min > max_size ? max_size : single_min
      
      if single_min && couple_min && max_size
        h[:$and] ||= []
        first = {
          single_parent: true,
          family_size: {:$gte => single_min}
        }
        first[:family_size][:$lte] = max_size if max_size < 12
        first[:family_size][:$ne] = 8 if rooms == 6
        second = {
          single_parent: {:$ne => true},
          family_size: {:$in => for_single_parent.map(&:to_i)}
        }
        h[:$and] << {:$or => [first, second]}
      end
      
      # revenues filter
      h[:monthly_revenues] = {:$gte => min_revenues} if min_revenues.present?
      
      # geo filter
      h[:truly_seeked_departments] = department if department.present?
      
      h[:$and] << {
        :$or => [
          {
            valid_seeked_types: typology
          },
          {
            valid_seeked_types: {:$eq => nil},
            typology_range: typology
          }
        ]
      }
      if (query = h.to_json) != eligibility_query && HousingOffer.published_statuses.include?(offer_status)
        
        self.eligibility_query = query
        
        eligible_applicant_ids = Hermes::HousingRequest.where(is_active: true, **h).pluck(:main_applicant_id).map(&:to_s)
        if eligible_applicant_ids.present?
          UpdateScoreJob.perform_later('housing_offer', eligible_applicant_ids, id.to_s)
        end
      end
    end
    
    def get_tax_threshold(cat = nil, for_floor = false)
      prefixe = for_floor ? 'floor_category_' : 'ceil_category_'
      send("#{prefixe}#{cat}")
    end
    
    class << self
      def rented_statuses
        RENTED_STATUSES
      end
      
      def published_statuses
        PUBLISHED_STATUSES
      end
      
      def search_by_postal_codes(postal_codes)
        weights = postal_codes.uniq.reverse.each_with_index.map do |postal_code, idx|
          {'$cond' => [{'$eq' => ['$postal_code', postal_code]}, (idx + 1), 0]}
        end
        project = fields.keys.map do |field|
          {field => 1}
        end.reduce({}, :merge)
        
        collection.aggregate([
                               {'$project' => project.merge('weight': {'$add': weights})},
                               {'$sort' => {'weight': -1}}
                             ])
      end
    end
    
    def ordered_pictures
      (ordered_picture_ids || []).map {|picture_id| pictures.find_by(id: picture_id)}
    end
    
    def add_picture(pic_id)
      self.ordered_picture_ids ||= []
      self.ordered_picture_ids << pic_id
      save
    end
    
    def remove_picture(pic_id)
      if ordered_picture_ids
        self.ordered_picture_ids.delete(pic_id)
        save
      end
    end
  
  
  end
end
